/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 */
package main

import (
	"context"
	"fmt"
	"io"
	"log"
	"os"
	"os/signal"
	"sort"
	"strconv"
	"strings"
	"time"

	"golang.org/x/sys/unix"
)

var LastErr error

func main() {
	log.SetOutput(os.Stderr)
	ctx, cancel := context.WithCancel(context.Background())
	hupChan := make(chan bool)         // apprise accStats of HUP
	sigChan := make(chan os.Signal, 1) // catch signals
	signal.Notify(
		sigChan,
		unix.SIGHUP,
		unix.SIGINT,
		unix.SIGTERM,
	)
	go func() {
		for {
			sig := <-sigChan
			switch sig {
			case unix.SIGINT, unix.SIGTERM:
				LastErr = fmt.Errorf("caught SIGINT/SIGTERM")
				cancel()
			case unix.SIGHUP:
				hupChan <- true
			}
		}
	}()
	if len(os.Args) != 2 {
		log.Fatalf("Usage: %s <log_file>\n", os.Args[0])
	}
	logPath := os.Args[1]
	lineChan := make(chan *string)  // tailFile -> accStats
	accChan := make(chan struct{})  // receive accChan done
	tailChan := make(chan struct{}) // receive tailChan done
	tasks := map[string](<-chan struct{}){
		"accstats": accChan,
		"tailfile": tailChan,
	}
	go accStats(cancel, lineChan, hupChan, accChan)
	go tailFile(ctx, cancel, logPath, lineChan, tailChan)
	os.Exit(mainLoop(ctx, tasks))
}

func mainLoop(
	ctx context.Context,
	tasks map[string](<-chan struct{}),
) int {
	for {
		select {
		case <-ctx.Done():
			exitCode := 0
			if LastErr != nil {
				log.Println(LastErr.Error())
				exitCode = 1
			}
			log.Println("terminating...")
			for k, v := range tasks {
				select {
				case <-v:
					log.Printf("stopped %s\n", k)
				case <-time.After(60 * time.Second):
					log.Printf("%s did not stop in time\n", k)
				}
			}
			return exitCode
		case <-time.After(100 * time.Millisecond):
			// do nothing; just throttling CPU
		}
	}
}

type Cursor struct {
	ctx   context.Context
	path  string
	file  *os.File
	buf   []byte
	pos   int64
	ino   uint64
	err   error
	lines int64
}

func (c *Cursor) Close() error {
	return c.file.Close()
}

func (c *Cursor) Open() error {
	return c.reopen()
}

func (c *Cursor) reopen() (err error) {
	_ = c.Close()
	c.file, err = os.Open(c.path)
	if err != nil {
		return fmt.Errorf("could not open file at '%s': %v", c.path, err)
	}
	stat := &unix.Stat_t{}
	err = unix.Stat(c.path, stat)
	if err != nil {
		return fmt.Errorf("unable to stat file '%s': %v", c.path, err)
	}
	c.ino = stat.Ino
	c.pos = 0
	c.buf = make([]byte, 0, 64*1024)
	return nil
}

func (c *Cursor) read() (line *string, err error) {
	for {
		i := 0
		for i < len(c.buf) {
			if c.buf[i] == '\n' {
				break
			}
			i++
		}
		if i != len(c.buf) {
			tmp := string(c.buf[:i+1])
			line = &tmp
			if i+1 < len(c.buf) { // line is not all of buffer
				copy(c.buf[:len(c.buf)-i-1], c.buf[i+1:])
			}
			c.buf = c.buf[:len(c.buf)-i-1]
			c.lines++
			return
		}
		if len(c.buf) == cap(c.buf) {
			err = fmt.Errorf("cursor: read: line buffer is too small")
			return
		}
		var n int
		n, err = c.file.Read(c.buf[len(c.buf):cap(c.buf)])
		if err != nil && err != io.EOF {
			return
		}
		c.buf = c.buf[:len(c.buf)+n]
		if n > 0 {
			continue
		}
		if err == io.EOF {
			stat := &unix.Stat_t{}
			err = unix.Stat(c.path, stat)
			if err != nil {
				c.err = fmt.Errorf("cursor: iter: unable to stat file '%s': %v", c.path, err)
				return
			}
			if stat.Ino != c.ino {
				c.err = c.reopen()
				if c.err != nil {
					return
				}
				continue
			}
			err = io.EOF
			return
		}
	}
}

func (c *Cursor) Iter() <-chan *string {
	ch := make(chan *string, 0)
	go func() {
		defer close(ch)
		for {
			select {
			case <-c.ctx.Done():
				log.Println("cursor: stopping iter...")
				return
			default:
			}
			line, err := c.read()
			if err == io.EOF {
				time.Sleep(100 * time.Millisecond)
				continue
			}
			if err != nil {
				c.err = err
				return
			}
			ch <- line
		}
	}()
	return ch
}

func (c *Cursor) LastErr() error {
	return c.err
}

func (c *Cursor) Lines() int64 {
	return c.lines
}

func NewCursor(ctx context.Context, path string) *Cursor {
	return &Cursor{
		ctx:  ctx,
		path: path,
	}
}

type cumulativeMean struct {
	mean    float64
	samples int64
}

func (m *cumulativeMean) AddSample(s int64) {
	m.samples++
	m.mean = m.mean + (float64(s)-m.mean)/float64(m.samples)
	return
}

func (m *cumulativeMean) Mean() float64 {
	return m.mean
}

func (m *cumulativeMean) Samples() int64 {
	return m.samples
}

func (m *cumulativeMean) Total() float64 {
	return m.mean * float64(m.samples)
}

func newCumulativeMean(d float64) *cumulativeMean {
	return &cumulativeMean{mean: d, samples: 1}
}

func accStats(
	cancel context.CancelFunc,
	ch <-chan *string,
	hupChan <-chan bool,
	doneChan chan<- struct{},
) {
	defer cancel()
	var e *string
	var more bool
	data := make(map[string]*cumulativeMean)
anFor:
	for {
		select {
		case <-hupChan:
			log.Println("caught HUP; printing stats...")
			printStats(data)
			continue
		case e, more = <-ch:
			if !more {
				log.Println("stopping accumulate...")
				break anFor
			}
		case <-time.After(100 * time.Millisecond):
			continue
		}
		parts := strings.Split(*e, " ")
		request := parts[5] + " " + parts[6]
		request = request[1:]
		procTimeUsecStr :=
			strings.TrimSuffix(parts[len(parts)-1], "\n")
		procTimeUsec, err :=
			strconv.ParseInt(procTimeUsecStr, 10, 64)
		if err != nil {
			log.Printf("WARN: acc stats: failed to parse serve time '%s' for log '%s'\n", parts[len(parts)-1], *e)
			continue
		}
		if _, ok := data[request]; !ok {
			data[request] =
				newCumulativeMean(float64(procTimeUsec))
			continue
		}
		data[request].AddSample(procTimeUsec)
	}
	printStats(data)
	doneChan <- struct{}{}
	return
}

func printStats(data map[string]*cumulativeMean) {
	sorted := make([]string, len(data))
	i := 0
	for k, _ := range data {
		sorted[i] = k
		i++
	}
	log.Println("sorting data by mean...")
	sort.Slice(sorted, func(i, j int) bool {
		di := data[sorted[i]].Mean()
		dj := data[sorted[j]].Mean()
		return di > dj
	})
	log.Println("printing data...")
	for _, k := range sorted {
		cm := data[k]
		time := cm.Total() * 1000.0 / float64(time.Second)
		fmt.Printf("mean=%fus; samples=%d; time=%fs; request=%s\n", cm.Mean(), cm.Samples(), time, k)
	}
	return
}

func tailFile(
	ctx context.Context,
	cancel context.CancelFunc,
	logPath string,
	ch chan<- *string,
	doneChan chan<- struct{},
) {
	defer close(ch)
	defer cancel()
	cur := NewCursor(ctx, logPath)
	err := cur.Open()
	if err != nil {
		log.Printf("tail file: unable to open cursor for path '%s': %v", logPath, err)
		return
	}
	lines := cur.Iter()
	var e *string
	for {
		select {
		case <-ctx.Done():
			log.Println("stopping tail...")
			log.Printf("processed %d logs\n", cur.Lines())
			doneChan <- struct{}{}
			return
		case e = <-lines:
			ch <- e
		case <-time.After(100 * time.Millisecond):
		}
	}
	if cur.LastErr() != nil {
		log.Println(cur.LastErr())
	}
	return
}
